;-----------------------------------------------------
; This code was mostly adapted from examples in
; The Undocumented PC, First edition, Frank van
; Gilluwe, Addison-Wesley, 1994
;-----------------------------------------------------
    .model small, c
    .386

    include support.asm

    ; The repeat count here and the value of
    ; TIMEOUT_COUNT could be too low for a
    ; fast CPU. 
    IO_DELAY MACRO
      REPEAT 10
        jmp $+2
      ENDM
    ENDM

    TIMEOUT_COUNT EQU 1000000

    .stack
    .data
    .code
    .startup

    call a20_enabled
    invoke w2dec, ax, ADDR outbuf
    print "A20 status = ",ADDR outbuf, NL

    call enable_a20

    ; In tests on a K5-PR166, after the call to
    ; enable_a20 the first call to a20_enabled
    ; always returned zero.    
    call a20_enabled
    invoke w2dec, ax, ADDR outbuf
    print "A20 status = ",ADDR outbuf, NL

    call a20_enabled
    invoke w2dec, ax, ADDR outbuf
    print "A20 status = ",ADDR outbuf, NL

    call a20_enabled
    invoke w2dec, ax, ADDR outbuf
    print "A20 status = ",ADDR outbuf, NL

    print NL, "Press any key to exit..."
    call waitkey
    .exit

;-----------------------------------------------------
; Returns with the carry flag clear and the value
; read from the keyboard controller in AL, or with
; the carry flag set if the status wait loop timed
; out. 
;-----------------------------------------------------
kb_read proc
    mov   ecx, TIMEOUT_COUNT
  @@:
    in    al, 64h       ; keyboard controller status
    IO_DELAY
    test  al, 1         ; output buffer status
    jnz   @F            ; 1 = byte available
    dec   ecx
    jnz   @B            ; try again
    stc
    jmp   fini
  @@:
    in    al, 60h       ; read byte
    clc    
  fini:
    ret
kb_read endp

;-----------------------------------------------------
; Writes the byte in AL to the keyboard controller
; data port. Returns with the carry flag set if
; the status wait loop timed out.
;-----------------------------------------------------
kb_write proc    
    mov   ecx, TIMEOUT_COUNT
  @@:
    push  ax
    in    al, 64h       ; keyboard controller status
    IO_DELAY
    test  al, 2         ; input buffer status
    pop   ax
    jz    @F            ; 0 = empty, OK to write
    dec   ecx
    jnz   @B            ; try again
    stc
    jmp   fini
  @@:
    out   60h, al       ; write byte
    clc
  fini:
    ret
kb_write endp

;-----------------------------------------------------
; Sends the command in AL to keyboard controller
; command port. Returns with the carry flag set if
; the status wait loop timed out.
;-----------------------------------------------------
kb_cmd proc
    mov   ecx, TIMEOUT_COUNT
  @@:
    push  ax
    in    al, 64h       ; keyboard controller status
    IO_DELAY
    test  al, 2         ; input buffer status
    pop   ax
    jz    @F            ; 0 = empty, OK to write
    dec   ecx
    jnz   @B            ; try again
    stc
    jmp   fini
  @@:
    out   64h, al       ; write byte
    IO_DELAY
    ; Wait for input buffer empty = command accepted.
    mov   ecx, TIMEOUT_COUNT
  @@:
    push  ax
    in    al, 64h       ; keyboard controller status
    test  al, 2         ; input buffer status
    pop   ax
    jz    @F            ; 0 = empty, OK to write
    dec   ecx
    jnz   @B            ; try again
    stc
    jmp   fini
  @@:
    clc
  fini:
    ret
kb_cmd endp

;----------------------------------------------------------
; Enables the A20 address line. Returns with the carry
; flag set if the status wait loop timed out.
;----------------------------------------------------------
.data
    errorID dw  0           ; *** debug only ***
.code
enable_a20 proc
    cli
    mov   errorID, 0        ; *** debug only *** 

    mov   al, 0d0h          ; Read Output Port
    call  kb_cmd
    jc    error_timeout

    inc   errorID           ; *** debug only ***

    call  kb_read           ; Get the current value
    jc    error_timeout

    or    al, 2             ; Set bit1 to enable A20
    
    inc   errorID           ; *** debug only ***

    push  ax
    mov   al, 0d1h          ; Write Output Port
    call  kb_cmd
    jc    error_timeout

    inc   errorID           ; *** debug only ***

    pop   ax
    call  kb_write          ; Write modified value back
    jc    error_timeout
    
    jmp   fini

  error_timeout:

    ; *** debug only ***
    invoke w2dec, errorID, ADDR outbuf
    print NL, "Timeout error ", ADDR outbuf, NL
    ; *** debug only ***

  fini:
    sti
    ret
enable_a20 endp

;----------------------------------------------------------
; Returns non-zero if the A20 address line is enabled.
; Assumes the A20 address line is disabled if the 
; address FFFF:10 wraps to the beginning of memory
; (so the words at 0:0 and FFFF:10 are identical).
;----------------------------------------------------------
a20_enabled proc uses ds es
    cli
    push  0ffffh
    pop   es
    mov   di, 10h           ; ES:DI = FFFF:10
    push  0
    pop   ds
    mov   si, 0             ; DS:SI = 0:0    
    not   WORD PTR ds:[si]  ; Modify value at 0:0
    xor   ax, ax
    cld
    cmpsw
    je    @F                ; Jump if A20 disabled
    inc   ax
  @@:
    not   WORD PTR ds:[si]  ; Restore value at 0:0
    sti
    ret
a20_enabled endp

end
